@@ -2,8 +2,10 @@ import inspect
 from enum import IntEnum
 from typing import Any, Callable, List, Mapping, NamedTuple, Set, Type, Union
 
-from .exceptions import ConfigError, Error, type_display
-from .validators import NoneType, find_validators, not_none_validator
+from .error_wrappers import ErrorWrapper
+from .errors import ConfigError, SequenceError
+from .utils import display_as_type
+from .validators import NoneType, dict_validator, find_validators, not_none_validator
 
 Required: Any = Ellipsis
 
@@ -72,7 +74,7 @@ class Field:
     @classmethod
     def infer(cls, *, name, value, annotation, class_validators, config):
         required = value == Required
-        field_config = _get_field_config(config, name)
+        field_config = config.get_field_config(name)
         return cls(
             name=name,
             type_=annotation,
@@ -86,7 +88,7 @@ class Field:
 
     def set_config(self, config):
         self.model_config = config
-        field_config = _get_field_config(config, self.name)
+        field_config = config.get_field_config(self.name)
         if field_config:
             self.alias = field_config.get('alias') or self.alias
             self.description = field_config.get('description') or self.description
@@ -113,7 +115,7 @@ class Field:
         self._populate_validators()
 
         self.info = {
-            'type': type_display(self.type_),
+            'type': display_as_type(self.type_),
             'default': self.default,
             'required': self.required,
         }
@@ -148,7 +150,7 @@ class Field:
                 default=self.default,
                 required=self.required,
                 allow_none=self.allow_none,
-                name=f'{self.name}_{type_display(t)}',
+                name=f'{self.name}_{display_as_type(t)}',
                 model_config=self.model_config,
             ) for t in types_]
         elif issubclass(origin, List):
@@ -209,85 +211,93 @@ class Field:
             ))
         return tuple(v)
 
-    def validate(self, v, values, index=None, cls=None):
+    def validate(self, v, values, *, loc, cls=None):
         if self.allow_none and v is None:
             return None, None
 
+        if not isinstance(loc, tuple):
+            loc = (loc,)
+
         if self.whole_pre_validators:
-            v, errors = self._apply_validators(v, values, index, cls, self.whole_pre_validators)
+            v, errors = self._apply_validators(v, values, loc, cls, self.whole_pre_validators)
             if errors:
                 return v, errors
 
         if self.shape is Shape.SINGLETON:
-            v, errors = self._validate_singleton(v, values, index, cls)
+            v, errors = self._validate_singleton(v, values, loc, cls)
         elif self.shape is Shape.MAPPING:
-            v, errors = self._validate_mapping(v, values, cls)
+            v, errors = self._validate_mapping(v, values, loc, cls)
         else:
             # list or set
-            v, errors = self._validate_sequence(v, values, cls)
+            v, errors = self._validate_sequence(v, values, loc, cls)
             if not errors and self.shape is Shape.SET:
                 v = set(v)
 
         if not errors and self.whole_post_validators:
-            v, errors = self._apply_validators(v, values, index, cls, self.whole_post_validators)
+            v, errors = self._apply_validators(v, values, loc, cls, self.whole_post_validators)
         return v, errors
 
-    def _validate_sequence(self, v, values, cls):
+    def _validate_sequence(self, v, values, loc, cls):
         result, errors = [], []
+
         try:
             v_iter = enumerate(v)
-        except TypeError as exc:
-            return v, Error(exc, None, None)
+        except TypeError:
+            return v, ErrorWrapper(SequenceError(), loc=loc, config=self.model_config)
+
         for i, v_ in v_iter:
-            single_result, single_errors = self._validate_singleton(v_, values, i, cls)
+            v_loc = *loc, i
+            single_result, single_errors = self._validate_singleton(v_, values, v_loc, cls)
             if single_errors:
                 errors.append(single_errors)
             else:
                 result.append(single_result)
+
         if errors:
             return v, errors
         else:
             return result, None
 
-    def _validate_mapping(self, v, values, cls):
-        if isinstance(v, dict):
-            v_iter = v
-        else:
-            try:
-                v_iter = dict(v)
-            except TypeError:
-                return v, Error(TypeError(f'value is not a valid dict, got {type_display(type(v))}'), None, None)
+    def _validate_mapping(self, v, values, loc, cls):
+        try:
+            v_iter = dict_validator(v)
+        except TypeError as exc:
+            return v, ErrorWrapper(exc, loc=loc, config=self.model_config)
 
         result, errors = {}, []
         for k, v_ in v_iter.items():
-            key_result, key_errors = self.key_field.validate(k, values, 'key', cls)
+            v_loc = *loc, '__key__'
+            key_result, key_errors = self.key_field.validate(k, values, loc=v_loc, cls=cls)
             if key_errors:
                 errors.append(key_errors)
                 continue
-            value_result, value_errors = self._validate_singleton(v_, values, k, cls)
+
+            v_loc = *loc, k
+            value_result, value_errors = self._validate_singleton(v_, values, v_loc, cls)
             if value_errors:
                 errors.append(value_errors)
                 continue
+
             result[key_result] = value_result
         if errors:
             return v, errors
         else:
             return result, None
 
-    def _validate_singleton(self, v, values, index, cls):
+    def _validate_singleton(self, v, values, loc, cls):
         if self.sub_fields:
             errors = []
             for field in self.sub_fields:
-                value, error = field.validate(v, values, index, cls)
+                value, error = field.validate(v, values, loc=loc, cls=cls)
                 if error:
                     errors.append(error)
                 else:
                     return value, None
-            return v, errors[0] if len(self.sub_fields) == 1 else errors
+            return v, errors
         else:
-            return self._apply_validators(v, values, index, cls, self.validators)
+            return self._apply_validators(v, values, loc, cls, self.validators)
 
-    def _apply_validators(self, v, values, index, cls, validators):
+    def _apply_validators(self, v, values, loc, cls, validators):
         for signature, validator in validators:
             try:
                 if signature is ValidatorSignature.JUST_VALUE:
@@ -300,7 +310,7 @@ class Field:
                     # ValidatorSignature.CLS_VALUE_KWARGS
                     v = validator(cls, v, values=values, config=self.model_config, field=self)
             except (ValueError, TypeError) as exc:
-                return v, Error(exc, self.type_, index)
+                return v, ErrorWrapper(exc, loc=loc, config=self.model_config)
         return v, None
 
     def __repr__(self):
@@ -314,12 +324,7 @@ class Field:
 
 
 def _get_validator_signature(validator):
-    try:
-        signature = inspect.signature(validator)
-    except ValueError:
-        # TODO we should probably have a white list of allowed validators here, rather than assuming
-        # happens on builtins like float
-        return ValidatorSignature.JUST_VALUE
+    signature = inspect.signature(validator)
 
     # bind here will raise a TypeError so:
     # 1. we can deal with it before validation begins
@@ -345,13 +350,6 @@ def _get_validator_signature(validator):
                           f'(cls, value) or (cls, value, *, values, config, field)') from e
 
 
-def _get_field_config(config, name):
-    field_config = config.fields.get(name)
-    if isinstance(field_config, str):
-        field_config = {'alias': field_config}
-    return field_config
-
-
 def _get_type_origin(obj):
     """
     Like obj.__class__ or type(obj) but for typing objects
