@@ -6,7 +6,8 @@ from pathlib import Path
 from types import FunctionType
 from typing import Any, Dict, Set, Type, Union
 
-from .exceptions import ConfigError, Error, Extra, Missing, ValidationError
+from .error_wrappers import ErrorWrapper, ValidationError
+from .errors import ConfigError, ExtraError, MissingError
 from .fields import Field, Validator
 from .parse import Protocol, load_file, load_str_bytes
 from .types import StrBytes
@@ -18,24 +19,32 @@ class BaseConfig:
     anystr_strip_whitespace = False
     min_anystr_length = 0
     max_anystr_length = 2 ** 16
-    min_number_size = -2 ** 64
-    max_number_size = 2 ** 64
     validate_all = False
     ignore_extra = True
     allow_extra = False
     allow_mutation = True
+    allow_population_by_alias = False
     use_enum_values = False
     fields = {}
     validate_assignment = False
+    error_msg_templates: Dict[str, str] = {}
+
+    @classmethod
+    def get_field_config(cls, name):
+        field_config = cls.fields.get(name) or {}
+        if isinstance(field_config, str):
+            field_config = {'alias': field_config}
+        return field_config
 
 
-def inherit_config(self_config, parent_config) -> Type[BaseConfig]:
+def inherit_config(self_config: Type[BaseConfig], parent_config: Type[BaseConfig]) -> Type[BaseConfig]:
     if not self_config:
-        return parent_config
-    for k, v in parent_config.__dict__.items():
-        if not (k.startswith('_') or hasattr(self_config, k)):
-            setattr(self_config, k, v)
-    return self_config
+        base_classes = parent_config,
+    elif self_config == parent_config:
+        base_classes = self_config,
+    else:
+        base_classes = self_config, parent_config
+    return type('Config', base_classes, {})
 
 
 TYPE_BLACKLIST = FunctionType, property, type, classmethod, staticmethod
@@ -59,7 +68,7 @@ class ValidatorGroup:
         if unused_validators:
             fn = ', '.join(unused_validators)
             raise ConfigError(f"Validators defined with incorrect fields: {fn} "
-                              f"(use check_fields=True if you're inheriting from the model and intended this)")
+                              f"(use check_fields=False if you're inheriting from the model and intended this)")
 
 
 def _extract_validators(namespace):
@@ -128,9 +137,7 @@ class MetaModel(ABCMeta):
         return super().__new__(mcs, name, bases, new_namespace)
 
 
-MISSING = Missing('field required')
-MISSING_ERROR = Error(MISSING, None, None)
-EXTRA_ERROR = Error(Extra('extra fields not permitted'), None, None)
+_missing = object()
 
 
 class BaseModel(metaclass=MetaModel):
@@ -156,9 +163,9 @@ class BaseModel(metaclass=MetaModel):
         elif not self.__config__.allow_mutation:
             raise TypeError(f'"{self.__class__.__name__}" is immutable and does not support item assignment')
         elif self.__config__.validate_assignment:
-            value_, error_ = self.fields[name].validate(value, self.dict(exclude={name}))
+            value_, error_ = self.fields[name].validate(value, self.dict(exclude={name}), loc=name)
             if error_:
-                raise ValidationError({name: error_})
+                raise ValidationError([error_])
             else:
                 self.__values__[name] = value_
         else:
@@ -183,7 +190,7 @@ class BaseModel(metaclass=MetaModel):
     def parse_obj(cls, obj):
         if not isinstance(obj, dict):
             exc = TypeError(f'{cls.__name__} expected dict not {type(obj).__name__}')
-            raise ValidationError([Error(exc, None, None)])
+            raise ValidationError([ErrorWrapper(exc, loc='__obj__')])
         return cls(**obj)
 
     @classmethod
@@ -196,7 +203,7 @@ class BaseModel(metaclass=MetaModel):
             obj = load_str_bytes(b, proto=proto, content_type=content_type, encoding=encoding,
                                  allow_pickle=allow_pickle)
         except (ValueError, TypeError, UnicodeDecodeError) as e:
-            raise ValidationError([Error(e, None, None)])
+            raise ValidationError([ErrorWrapper(e, loc='__obj__')])
         return cls.parse_obj(obj)
 
     @classmethod
@@ -254,23 +261,28 @@ class BaseModel(metaclass=MetaModel):
 
     def _process_values(self, input_data: dict) -> Dict[str, Any]:  # noqa: C901 (ignore complexity)
         values = {}
-        errors = {}
+        errors = []
 
         for name, field in self.__fields__.items():
-            value = input_data.get(field.alias, MISSING)
-            if value is MISSING:
+            value = input_data.get(field.alias, _missing)
+            if value is _missing and self.__config__.allow_population_by_alias and field.alt_alias:
+                value = input_data.get(field.name, _missing)
+
+            if value is _missing:
                 if self.__config__.validate_all or field.validate_always:
-                    value = field.default
+                    value = deepcopy(field.default)
                 else:
                     if field.required:
-                        errors[field.alias] = MISSING_ERROR
+                        errors.append(ErrorWrapper(MissingError(), loc=field.alias, config=self.__config__))
                     else:
-                        values[name] = field.default
+                        values[name] = deepcopy(field.default)
                     continue
 
-            v_, errors_ = field.validate(value, values, cls=self.__class__)
-            if errors_:
-                errors[field.alias] = errors_
+            v_, errors_ = field.validate(value, values, loc=field.alias, cls=self.__class__)
+            if isinstance(errors_, ErrorWrapper):
+                errors.append(errors_)
+            elif isinstance(errors_, list):
+                errors.extend(errors_)
             else:
                 values[name] = v_
 
@@ -283,7 +295,7 @@ class BaseModel(metaclass=MetaModel):
                 else:
                     # config.ignore_extra is False
                     for field in sorted(extra):
-                        errors[field] = EXTRA_ERROR
+                        errors.append(ErrorWrapper(ExtraError(), loc=field, config=self.__config__))
 
         if errors:
             raise ValidationError(errors)
