@@ -1,12 +1,16 @@
 import re
+from contextlib import contextmanager
 from importlib import import_module
-from typing import Tuple
+from typing import Tuple, _TypingBase
+
+from . import errors
 
 try:
     import email_validator
 except ImportError:
     email_validator = None
 
+
 PRETTY_REGEX = re.compile(r'([\w ]*?) *<(.*)> *')
 
 
@@ -29,8 +33,14 @@ def validate_email(value) -> Tuple[str, str]:
         name, value = m.groups()
     else:
         name = None
+
     email = value.strip()
-    email_validator.validate_email(email, check_deliverability=False)
+
+    try:
+        email_validator.validate_email(email, check_deliverability=False)
+    except email_validator.EmailNotValidError as e:
+        raise errors.EmailError() from e
+
     return name or email[:email.index('@')], email.lower()
 
 
@@ -102,3 +112,22 @@ def truncate(v, *, max_len=80):
     if len(v) > max_len:
         v = v[:max_len - 1] + '…'
     return v
+
+
+def display_as_type(v):
+    if not isinstance(v, _TypingBase) and not isinstance(v, type):
+        v = type(v)
+
+    try:
+        return v.__name__
+    except AttributeError:
+        # happens with unions
+        return str(v)
+
+
+@contextmanager
+def change_exception(raise_exc, *except_types):
+    try:
+        yield
+    except except_types as e:
+        raise raise_exc from e
